import ast
from typing import Literal, Optional

import pandas as pd
import torch
from torch.utils.data import Dataset, DataLoader
from transformers import RobertaTokenizer, PreTrainedTokenizer, DataCollatorForLanguageModeling

from train.mask import mask_conditions

TOKENIZER: PreTrainedTokenizer = RobertaTokenizer.from_pretrained('Salesforce/codet5-small')

BATCH_SIZE: int = 8
MAX_TOKEN_LENGTH: int = 512

# Probability of masking a token during pretraining
PRETRAIN_MLM_PROB: float = 0.15


class PythonCodeDataset(Dataset):
    """Dataset class for pre-training"""

    def __init__(self, tokenizer, dataframe, max_len):
        self.tokenizer = tokenizer
        self.data = dataframe
        self.max_len = max_len

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        code = self.data.iloc[index]['source']
        inputs = self.tokenizer.encode_plus(code, None, add_special_tokens=True, max_length=self.max_len,
                                            padding='max_length', truncation=True)
        return {'input_ids': torch.tensor(inputs['input_ids'], dtype=torch.long),
                'attention_mask': torch.tensor(inputs['attention_mask'], dtype=torch.long)}


def build_pretrain_dataloader(pretrain_df: pd.DataFrame) -> DataLoader:
    pretrain_dataset = PythonCodeDataset(TOKENIZER, pretrain_df, MAX_TOKEN_LENGTH)
    data_collator = DataCollatorForLanguageModeling(tokenizer=TOKENIZER, mlm=True, mlm_probability=PRETRAIN_MLM_PROB)
    return DataLoader(pretrain_dataset, batch_size=BATCH_SIZE, shuffle=True, collate_fn=data_collator)


class MaskedIfDataset(PythonCodeDataset):
    """Fine-tuning and evaluation dataset classes"""

    def __getitem__(self, index):
        masked_code = self.data.iloc[index]['masked_code']
        ground_truth = self.data.iloc[index]['ground_truth']
        inputs = self.tokenizer(masked_code, max_length=self.max_len, padding='max_length', truncation=True,
                                return_tensors="pt")
        labels = self.tokenizer(ground_truth, max_length=self.max_len, padding='max_length', truncation=True,
                                return_tensors="pt").input_ids
        labels[labels == self.tokenizer.pad_token_id] = -100
        return {'input_ids': inputs.input_ids.squeeze(),
                'attention_mask': inputs.attention_mask.squeeze(),
                'labels': labels.squeeze()}


Kind = Literal['train'] | Literal['val'] | Literal['test'] | Literal['test_usi']


def build_fine_tune_dataloader(df: pd.DataFrame, kind: Kind) -> DataLoader:
    print(f"{kind} dataset: {len(df.index)} functions found")
    df = mask_conditions(df, kind)
    print(f"{kind} dataset: {len(df.index)} conditions found")

    dataset = MaskedIfDataset(TOKENIZER, df, MAX_TOKEN_LENGTH)
    return DataLoader(dataset, batch_size=BATCH_SIZE, shuffle=kind == 'train')


def decode_tokenized(output) -> Optional[str]:
    label_trimmed = [token for token in output if token != TOKENIZER.pad_token_id and token != -100]
    if label_trimmed:
        return TOKENIZER.decode(label_trimmed, skip_special_tokens=True)
    else:
        return None
