import pandas as pd
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from train.dataset import decode_tokenized, TOKENIZER
from train.mask import normalize_condition


def compute_accuracy(outputs, batch, track_predictions=False, confidences=None) -> tuple[int, int, list[dict]]:
    correct_predictions, total_predictions = 0, 0

    decoded_outputs = [TOKENIZER.decode(output, skip_special_tokens=True) for output in outputs]

    if track_predictions:
        confidences = confidences.tolist()
    else:
        confidences = [None] * len(decoded_outputs)

    tracking = []

    # Decode labels with added check for None values
    decoded_labels = []
    for label in batch['labels']:
        decoded_labels.append(decode_tokenized(label))

    # Calculate accuracy
    for output, label, confidence, in_ids in zip(decoded_outputs, decoded_labels, confidences, batch['input_ids']):
        is_correct = None
        if label is not None:
            total_predictions += 1
            is_correct = False

            output = normalize_condition(output)
            label = normalize_condition(label)

            if output == label:
                correct_predictions += 1
                is_correct = True

        if track_predictions:
            tracking.append({
                'input': TOKENIZER.decode(in_ids) \
                    .replace("<pad>", "") \
                    .replace("<s>", "") \
                    .replace("</s>", ""),
                'is_correct': is_correct,
                'expected_cond': label,
                'predicted_cond': output,
                'score': confidence
            })

    return correct_predictions, total_predictions, tracking


def evaluate_accuracy(model, dataloader: DataLoader, device, track_predictions=False) -> tuple[
    float, float, list[dict]]:
    """Returns the accuracy and loss on the given validation set"""

    model.eval()
    total_loss = 0

    correct_predictions, total_predictions = 0, 0

    tracking = []

    for batch in tqdm(dataloader, desc="Evaluating"):
        batch = {k: v.to(device) for k, v in batch.items()}
        with torch.no_grad():
            inputs = {'input_ids': batch['input_ids'], 'attention_mask': batch['attention_mask'],
                      'labels': batch['labels']}
            outputs = model(**inputs)
            generation = model.generate(batch['input_ids'], attention_mask=batch['attention_mask'], max_length=512)

        # Compute prediction score as inverse entropy of the T5 model logits. If their softmax has values close to 0 or
        # 1, confidence is high
        loss, logits = outputs[:2]

        logits_flattened = outputs.logits.flatten(1, 2)
        probabilities = torch.nn.functional.softmax(logits_flattened, dim=1)
        entropy = -torch.sum(probabilities * torch.log(probabilities), dim=1)
        confidence = 1.0 - entropy / torch.log(torch.tensor(probabilities.size(dim=1)))

        total_loss += loss.item()

        c_pred, total_pred, t = compute_accuracy(generation, batch, track_predictions, confidence)
        correct_predictions += c_pred
        total_predictions += total_pred
        tracking.extend(t)

    return correct_predictions / total_predictions if total_predictions > 0 else 0, total_loss / len(
        dataloader), tracking
