import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from train.evaluate import evaluate_accuracy, compute_accuracy

ADAM_LR: float = 5e-5
EARLY_STOPPING_PATIENCE: int = 3
PLOT_FILENAME: str = "training_metrics.png"


def fine_tune_with_eval(model, device, train_loader: DataLoader, val_loader: DataLoader, epochs: int,
                        save_dir: str) -> int:
    optimizer = torch.optim.AdamW(model.parameters(), lr=ADAM_LR)
    best_epoch = 0
    best_eval_accuracy = 0
    patience_counter = 0
    train_accuracies, train_losses, val_accuracies, val_losses = [], [], [], []

    print(f"Tuning for {epochs} epochs")

    max_epoch = 0

    for epoch in range(epochs):
        max_epoch = epoch

        model.train()
        total_loss = 0

        correct_predictions, total_predictions = 0, 0
        train_accuracy = 0

        train_dl = tqdm(train_loader, desc=f"Train E{epoch + 1}/{epochs}")
        i = 0

        # Training loop with tqdm for progress tracking
        for batch in train_dl:
            i += 1
            batch = {k: v.to(device) for k, v in batch.items()}
            optimizer.zero_grad()
            inputs = {'input_ids': batch['input_ids'], 'attention_mask': batch['attention_mask'],
                      'labels': batch['labels']}
            outputs = model(**inputs)
            generation = model.generate(batch['input_ids'], attention_mask=batch['attention_mask'], max_length=512)

            loss = outputs.loss
            total_loss += loss.item()
            loss.backward()
            optimizer.step()

            # Update train accuracy
            c_pred, total_pred, _ = compute_accuracy(generation, batch)
            correct_predictions += c_pred
            total_predictions += total_pred
            train_accuracy = correct_predictions / total_predictions if total_predictions > 0 else 0

            train_dl.set_description(f"Train E{epoch + 1}/{epochs} loss={total_loss / i:.6f} acc={train_accuracy:.4f}")

        average_loss = total_loss / len(train_loader)
        train_losses.append(average_loss)
        train_accuracies.append(train_accuracy)

        # Evaluate on the validation set
        val_accuracy, val_loss, _ = evaluate_accuracy(model, val_loader, device)
        val_accuracies.append(val_accuracy)
        val_losses.append(val_loss)
        print(f"E{epoch + 1}/{epochs}, Train Loss: {average_loss:.4f}, Train Accuracy: {train_accuracy:.4f}, "
              f"Val Loss: {val_loss:.4f}, Val Accuracy: {val_accuracy:.4f}")

        # Early stopping and checkpointing
        if val_accuracy > best_eval_accuracy:
            best_eval_accuracy = val_accuracy
            best_epoch = epoch
            patience_counter = 0
        else:
            patience_counter += 1
            if patience_counter >= EARLY_STOPPING_PATIENCE:
                print("Early stopping triggered.")
                break

        save_directory = os.path.join(save_dir, str(epoch))
        model.save_pretrained(save_directory)

    df = pd.DataFrame({'train_loss': train_losses,
                       'train_acc': train_accuracies,
                       'val_loss': val_losses,
                       'val_acc': val_accuracies},
                      index=list(range(1, max_epoch + 2)))

    df.to_csv(os.path.join(save_dir, "stats.csv"))

    return best_epoch


def plot_loss_acc(train_losses: list[float], val_losses: list[float], train_accuracies: list[float],
                  val_accuracies: list[float], save_path: str):
    plt.figure(figsize=(12, 10))

    plt.subplot(2, 2, 1)
    plt.plot(train_losses, label='Training Loss')
    plt.title('Training Loss')
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.xticks(np.arange(0, 20, step=1))
    plt.yticks(np.arange(0, 1.4, step=0.2))
    plt.legend()

    plt.subplot(2, 2, 2)
    plt.plot(val_losses, label='Validation Loss', color='orange')
    plt.title('Validation Loss')
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.xticks(np.arange(0, 20, step=1))
    plt.yticks(np.arange(0, 1.4, step=0.2))
    plt.legend()

    plt.subplot(2, 2, 3)
    plt.plot(train_accuracies, label='Training Accuracy', color='green')
    plt.title('Training Accuracy')
    plt.xlabel('Epoch')
    plt.ylabel('Accuracy')
    plt.xticks(np.arange(0, 20, step=1))
    plt.yticks(np.arange(0, 1.1, step=0.1))
    plt.legend()

    plt.subplot(2, 2, 4)
    plt.plot(val_accuracies, label='Validation Accuracy', color='red')
    plt.title('Validation Accuracy')
    plt.xlabel('Epoch')
    plt.ylabel('Accuracy')
    plt.xticks(np.arange(0, 20, step=1))
    plt.yticks(np.arange(0, 1.1, step=0.1))
    plt.legend()

    plt.tight_layout()

    plt.savefig(os.path.join(save_path, PLOT_FILENAME))
