from dataclasses import dataclass
from io import BytesIO
from tokenize import tokenize

import pandas as pd
import swifter
from sklearn.model_selection import train_test_split
from train.mask import MASK_TOKEN
from tqdm import tqdm

# PreTrain/Train = 1-FINE_TUNE_FRAC/FINE_TUNE_FRAC split
FINE_TUNE_FRAC: float = 0.5

# Splits for the fine tune dataset
# TRAIN_FRAC = 1-VAL_FRAC-TEST_FRAC
TEST_FRAC: float = 0.1
VAL_FRAC: float = 0.1

LOW_PERCENTILE: float = 0.25
HIGH_PERCENTILE: float = 0.75


@dataclass
class DataSet:
    pretrain_df: pd.DataFrame
    fine_tune_train_df: pd.DataFrame
    fine_tune_val_df: pd.DataFrame
    fine_tune_test_df: pd.DataFrame
    usi_test_df: pd.DataFrame

    def __init__(self, pretrain_df: pd.DataFrame, fine_tune_train_df: pd.DataFrame, fine_tune_val_df: pd.DataFrame,
                 fine_tune_test_df: pd.DataFrame, usi_test_df: pd.DataFrame):
        self.pretrain_df = pretrain_df
        self.fine_tune_train_df = fine_tune_train_df
        self.fine_tune_val_df = fine_tune_val_df
        self.fine_tune_test_df = fine_tune_test_df
        self.usi_test_df = usi_test_df

    @staticmethod
    def load(dataset_path: str, usi_test_dataset_path: str, random_state: int):
        df = pd.read_parquet(dataset_path)
#        df = df.iloc[0:100, :]  # debug
        df = filter_outliers(df)

        pretrain_df, fine_tune_df = train_test_split(df, test_size=FINE_TUNE_FRAC, random_state=random_state)
        val_df = fine_tune_df.sample(frac=VAL_FRAC, random_state=random_state)
        test_df = fine_tune_df.drop(val_df.index).sample(frac=TEST_FRAC / (1 - VAL_FRAC), random_state=random_state)
        fine_tune_df = fine_tune_df.drop(val_df.index).drop(test_df.index)
        usi_test_df = pd.read_csv(usi_test_dataset_path, index_col=0)
#        usi_test_df = usi_test_df.iloc[0:10, :] # debug

        print(f"pretrain dataset: {len(pretrain_df.index)} functions loaded")
        print(f"finetune train dataset: {len(fine_tune_df.index)} functions loaded")
        print(f"finetune val dataset: {len(val_df.index)} functions loaded")
        print(f"finetune test dataset: {len(test_df.index)} functions loaded")

        return DataSet(pretrain_df, fine_tune_df, val_df, test_df, usi_test_df)


def filter_outliers(df: pd.DataFrame) -> pd.DataFrame:
    assert swifter is not None

    def count_tokens(s: str) -> int:
        try:
            count = 0
            for _ in tokenize(BytesIO(s.encode('utf-8')).readline):
                count += 1
            return count
        except:
            return 0

    df["token_length"] = df["source"].swifter.apply(count_tokens)

    low_qty = df["token_length"].quantile(LOW_PERCENTILE)
    mask_low = df["token_length"] < low_qty
    print(f"Excluding token lengths < {low_qty} ({LOW_PERCENTILE * 100:02.02f}%ile): {sum(mask_low)} instances")

    high_qty = df["token_length"].quantile(HIGH_PERCENTILE)
    mask_high = df["token_length"] > high_qty
    print(f"Excluding token lengths > {high_qty} ({HIGH_PERCENTILE * 100:02.02f}%ile): {sum(mask_high)} instances")

    return df[~mask_high & ~mask_low]
