import ast
import sys
from _ast import Load
from dataclasses import dataclass

import pandas as pd
from pandas import Series
from tqdm import tqdm

MASK_TOKEN: str = "<mask>"


def strip_parentheses(input_string):
    input_string = input_string.strip()
    if input_string.startswith('(') and input_string.endswith(')'):
        # Dyck Language algorithm

        paren_count: int = 0
        for char in input_string:
            if char == '(':
                paren_count += 1
            elif char == ')':
                if paren_count == 0:
                    return input_string  # closing a non-matching open paren
                else:
                    paren_count -= 1

        if paren_count == 0:
            return input_string[1:-1]  # strip if parens are balanced

    return input_string


@dataclass
class FineTrainInstance:
    masked_function: str
    condition: str

    def __init__(self, masked_function: str, condition: str):
        self.masked_function = masked_function
        self.condition = condition

    @staticmethod
    def from_function(function: str) -> list['FineTrainInstance']:
        try:
            tree = ast.parse(function)
        except SyntaxError:
            return []

        instances: list['FineTrainInstance'] = []

        for t in ast.walk(tree):
            if isinstance(t, ast.If):
                # swap in place the condition with the mask node
                cond = t.test

                # Replace the condition with a mask node and build the masked function source
                # treat "<mask>" as a variable rvalue, which may not ever happen in real source code as it is not a
                # valid identifier. However, this makes ast.unparse happily print "<mask>" as a string
                t.test = ast.Name(identifier_id=MASK_TOKEN, expr_context_ctx=Load())
                t.test.id = MASK_TOKEN
                masked_fun = ast.unparse(tree)

                instances.append(FineTrainInstance(masked_fun, strip_parentheses(ast.unparse(cond))))

                # restore the condition
                t.test = cond

        return instances


def normalize_condition(c: str) -> str:
    c = c.strip()

    try:
        # reformat if syntax is parsable, otherwise return as-is
        return strip_parentheses(ast.unparse(ast.parse(c)))
    except SyntaxError:
        return c


def mask_conditions(df_source: pd.DataFrame, kind: str) -> pd.DataFrame:
    if kind != 'test_usi':
        df = pd.DataFrame(columns=['masked_code', 'ground_truth'])
        instances = df_source["source"].swifter.apply(lambda s: FineTrainInstance.from_function(s))

        i = 0
        for row in tqdm(instances, desc=f"Building {kind}", total=len(df_source.index)):
            for instance in row:
                df.loc[i, 'masked_code'] = instance.masked_function
                df.loc[i, 'ground_truth'] = instance.condition
                i += 1
    else:
        df = pd.DataFrame(columns=['masked_code', 'ground_truth'], index=df_source.index)

        def canonicalize(c: str) -> pd.Series:
            prefixes = ["if ", "elif "]
            found_prefix = ""
            postfix = ":"

            c = c.strip()

            for prefix in prefixes:
                if c.startswith(prefix):
                    c = c[len(prefix):]
                    found_prefix = prefix
                    break

            if c.endswith(postfix):
                c = c[:len(c) - len(postfix)]

            c = normalize_condition(c)

            return pd.Series([found_prefix, c], index=['found_prefix', 'c'])

        # Canonicalize condition string
        df[['prefix', 'ground_truth']] = df_source['target_block'].swifter.apply(canonicalize)
        df['masked_code'] = df_source['input_method'].copy()

        # Our model is only able to predict the if condition itself, so we re-inject the "if"/"elif" and ":" token
        # back in the input
        df['masked_code'] = df[['prefix', 'masked_code']] \
            .apply(lambda s: s['masked_code'].replace("<fill-in>", s['prefix'] + " " + MASK_TOKEN + " :"), axis=1)

    return df
